@@ -35,8 +35,6 @@ import matplotlib.pyplot as plt
 from typing import Tuple, Dict, List, Union, Set, Callable, Mapping, Sequence
 from decimal import Decimal
 
-np.seterr(all="raise")
-
 
 class CollapsedTree:
     r"""A collapsed tree, modeled as an infinite type Galton-Watson process run
@@ -46,9 +44,9 @@ class CollapsedTree:
         tree: :class:`ete3.TreeNode` object with ``abundance`` node features
 
     Args:
-        tree: ete3 tree with ``abundance`` node features. Nonzero abundances expected on leaves,
-            and optionally nodes adjacent to leaves via a path of length zero edges. If uncollapsed,
-            it will be collapsed along branches with no mutations. Can be ommitted on initializaion, and later simulated.
+        tree: ete3 tree with ``abundance`` node features. If uncollapsed,
+            it will be collapsed along branches with no mutations.
+            Can be ommitted on initializaion, and later simulated.
             If a tree is provided, names of nodes with abundance 0 will not be preserved.
         allow_repeats: tolerate the existence of nodes with the same genotype after collapse, e.g. in sister clades.
     """
@@ -60,6 +58,16 @@ class CollapsedTree:
             self.tree = tree.copy()
             self.tree.dist = 0
 
+            def merge_isotype_dicts(parent_isotype, child_isotype):
+                # values are abundances and keys are isotypes.
+                parent_isotype = dict(parent_isotype)
+                for key, val in child_isotype.items():
+                    if key in parent_isotype:
+                        parent_isotype[key] = max(parent_isotype[key], val)
+                    else:
+                        parent_isotype[key] = val
+                return frozendict(parent_isotype)
+
             # remove unobserved internal unifurcations
             for node in self.tree.iter_descendants():
                 if node.abundance == 0 and len(node.children) == 1:
@@ -75,12 +83,24 @@ class CollapsedTree:
             # length if the node is a leaf and it's parent has nonzero
             # abundance we combine taxa names to a set to acommodate
             # bootstrap samples that result in repeated genotypes
-            observed_genotypes = set((leaf.name for leaf in self.tree))
+            observed_genotypes = set(
+                node.name for node in self.tree.traverse() if node.abundance
+            )
             observed_genotypes.add(self.tree.name)
             for node in self.tree.get_descendants(strategy="postorder"):
                 if node.dist == 0:
+                    # if an abundance is nonzero, that's the right one.
                     node.up.abundance = max(node.abundance, node.up.abundance)
-                    node.up.isotype = node.isotype
+                    # isotype is dictionary with isotype as key and observed
+                    # abundance as value
+                    if "isotype" in node.features:
+                        node.up.isotype = merge_isotype_dicts(
+                            node.up.isotype, node.isotype
+                        )
+                    # original_ids is a set of observed ids corresponding to
+                    # each node
+                    if "original_ids" in node.features:
+                        node.up.original_ids = node.original_ids | node.up.original_ids
                     if isinstance(node.name, str):
                         node_set = set([node.name])
                     else:
@@ -100,13 +120,19 @@ class CollapsedTree:
                             node.up.name = node.up.name[0]
                     node.delete(prevent_nondicotomic=False)
 
+                if "isotype" in node.features:
+                    node.add_feature(
+                        "inferred_isotype", min(node.isotype.keys(), default=None)
+                    )
+            if "isotype" in node.features:
+                self.tree.add_feature(
+                    "inferred_isotype", min(self.tree.isotype.keys(), default=None)
+                )
+
             final_observed_genotypes = set()
             for node in self.tree.traverse():
                 if node.abundance > 0 or node == self.tree:
-                    for name in (
-                        (node.name,) if isinstance(node.name, str) else node.name
-                    ):
-                        final_observed_genotypes.add(name)
+                    final_observed_genotypes.add(node.name)
             if final_observed_genotypes != observed_genotypes:
                 raise RuntimeError(
                     "observed genotypes don't match after "
@@ -344,6 +370,7 @@ class CollapsedTree:
 
         return (logf_result, np.array([dlogfdp_result, dlogfdq_result]))
 
+    @np.errstate(all="raise")
     def ll(
         self,
         p: np.float64,
@@ -657,7 +684,7 @@ class CollapsedTree:
         Args:
             file_name: file name (.nk suffix recommended)
         """
-        self.tree.write(format=1, outfile=file_name)
+        self.tree.write(format=1, outfile=file_name, format_root_node=True)
 
     def compare(
         self, tree2: CollapsedTree, method: str = "identity"
@@ -824,7 +851,7 @@ class CollapsedTree:
                     compatibility_ += weights[i] if weights is not None else 1
             node.support = compatibility_ if compatibility else support
 
-    def local_branching(self, tau=1, tau0=0.1):
+    def local_branching(self, tau=1, tau0=0.1, infinite_root_branch=True):
         r"""Add local branching statistics (Neher et al. 2014) as tree node
         features to the ETE tree attribute.
         After execution, all nodes will have new features ``LBI``
@@ -834,45 +861,48 @@ class CollapsedTree:
         Args:
             tau: decay timescale for exponential filter
             tau0: effective branch length for branches with zero mutations
+            infinite_root_branch: calculate assuming the root node has an infinite branch
         """
         # the fixed integral contribution for clonal cells indicated by abundance annotations
         clone_contribution = tau * (1 - np.exp(-tau0 / tau))
 
-        # post-order traversal to populate downward integral for each node
+        # post-order traversal to populate downward integrals for each node
         for node in self.tree.traverse(strategy="postorder"):
             if node.is_leaf():
-                node.add_feature(
-                    "LB_down",
-                    node.abundance * clone_contribution if node.abundance > 1 else 0,
-                )
+                node.LB_down = {
+                    node: node.abundance * clone_contribution
+                    if node.abundance > 1
+                    else 0
+                }
             else:
-                node.add_feature(
-                    "LB_down",
-                    node.abundance * clone_contribution
-                    + sum(
-                        tau * (1 - np.exp(-child.dist / tau))
-                        + np.exp(-child.dist / tau) * child.LB_down
-                        for child in node.children
-                    ),
-                )
+                node.LB_down = {node: node.abundance * clone_contribution}
+                for child in node.children:
+                    node.LB_down[child] = tau * (
+                        1 - np.exp(-child.dist / tau)
+                    ) + np.exp(-child.dist / tau) * sum(child.LB_down.values())
 
         # pre-order traversal to populate upward integral for each node
         for node in self.tree.traverse(strategy="preorder"):
             if node.is_root():
-                node.add_feature("LB_up", 0)
+                # integral corresponding to infinite branch above root node
+                node.LB_up = tau if infinite_root_branch else 0
             else:
-                node.add_feature(
-                    "LB_up",
-                    tau * (1 - np.exp(-node.dist / tau))
-                    + np.exp(-node.dist / tau) * (node.up.LB_up + node.up.LB_down),
+                node.LB_up = tau * (1 - np.exp(-node.dist / tau)) + np.exp(
+                    -node.dist / tau
+                ) * (
+                    node.up.LB_up
+                    + sum(
+                        node.up.LB_down[message]
+                        for message in node.up.LB_down
+                        if message != node
+                    )
                 )
 
-        # finally, compute LBI (LBR) as the sum (ratio) of upward and downward integrals at each node
+        # finally, compute LBI (LBR) as the sum (ratio) of downward and upward integrals at each node
         for node in self.tree.traverse():
-            node.add_feature("LBI", node.LB_down + node.LB_up)
-            node.add_feature(
-                "LBR", node.LB_down / node.LB_up if not node.is_root() else np.nan
-            )
+            node_LB_down_total = sum(node.LB_down.values())
+            node.LBI = node_LB_down_total + node.LB_up
+            node.LBR = node_LB_down_total / node.LB_up
 
 
 def _requires_dag(func):
@@ -899,21 +929,15 @@ class CollapsedForest:
 
     Args:
         forest: list of :class:`ete3.Tree`
-        sequence_counts: a mapping of observed sequences to observed abundances
     """
 
     def __init__(
         self,
         forest: List[Union[CollapsedTree, ete3.Tree]] = None,
-        sequence_counts: Mapping[str, int] = None,
     ):
         if forest is not None:
             if len(forest) == 0:
                 raise ValueError("passed empty tree list")
-            if sequence_counts is None:
-                raise ValueError(
-                    "an abundance dictionary must be provided to the keyword argument sequence_counts"
-                )
             if isinstance(forest[0], CollapsedTree):
                 forest = [ctree.tree for ctree in forest]
             elif not isinstance(forest[0], ete3.Tree):
@@ -921,22 +945,26 @@ class CollapsedForest:
                     "If provided, `forest` argument should contain a list of ete3 trees or CollapsedTrees."
                 )
             # Collect stats for validation
-            model_tree = disambiguate(forest[0].copy())
-            # include root in counts
+            model_tree = forest[0].copy()
+            leaf_seqs = {node.sequence: node.name for node in model_tree.get_leaves()}
+            root_seq = model_tree.sequence
+            model_tree = disambiguate(model_tree)
+            # include root in counts, since deduplicate will never let observed
+            # root be a leaf.
             counts = {node.name: node.abundance for node in model_tree.iter_leaves()}
+            counts[model_tree.name] = model_tree.abundance
             self._validation_stats = {
                 "counts": counts,
                 "root": model_tree.name,
                 "parsimony_score": sum([node.dist for node in model_tree.traverse()]),
-                "root_seq": model_tree.sequence,
-                "leaf_seqs": {
-                    node.sequence: node.name for node in model_tree.get_leaves()
-                },
+                "root_seq": root_seq,
             }
+            if not any(_is_ambiguous(key) for key in leaf_seqs):
+                self._validation_stats["leaf_seqs"] = leaf_seqs
             # Making this a private variable so that trying to access forest
             # attribute as before won't just give a confusing type or
             # attribute error.
-            self._forest = _make_dag(forest, sequence_counts=sequence_counts)
+            self._forest = _make_dag(forest)
             self.n_trees = self._forest.count_trees()
         else:
             self._forest = None
@@ -964,6 +992,7 @@ class CollapsedForest:
             coll.Counter([tree._cm_counts for tree in self._ctrees]).items()
         )
 
+    @np.errstate(all="raise")
     def ll(
         self,
         p: np.float64,
@@ -1127,10 +1156,11 @@ class CollapsedForest:
             if verbose:
                 print("Isotype parsimony will be used as a ranking criterion")
             # Check for missing isotype data in all but root node, and fake root-adjacent leaf node
+            rootname = list(self._forest.dagroot.children())[0].attr["name"]
             if any(
                 not node.attr["isotype"]
                 for node in self._forest.preorder()
-                if not node.is_root() and node.attr["name"] != ""
+                if not node.is_root() and node.attr["name"] != rootname
             ):
                 warnings.warn(
                     "Some isotype data seems to be missing. Isotype parsimony scores may be incorrect."
@@ -1420,47 +1450,66 @@ class CollapsedForest:
         Returns:
             :meth:`CollapsedTree` object matching the topology of ``clade_tree``, but fully collapsed.
         """
+        # Allow only leaf nodes to have nonzero abundance, so that ctree
+        # collapsing takes care of all uncollapsed leaves. This works for
+        # observed root because all trees have added pseudo leaf with root name
+        # and abundance.
+        # original_ids is a set of original leaf node ids, or empty if node is
+        # not a leaf.
         etetree = clade_tree.to_ete(
             name_func=lambda n: n.attr["name"],
             features=["sequence"],
             feature_funcs={
-                "abundance": lambda n: n.attr["abundance"],
+                "abundance": lambda n: n.label.abundance if n.is_leaf() else 0,
                 "isotype": lambda n: n.attr["isotype"],
+                "original_ids": lambda n: n.attr["original_ids"],
             },
         )
 
-        # Remove dummy leaf added below root for hDAG compatibility
-        dummyleaves = [
-            node for node in etetree.children if node.is_leaf() and node.abundance == 0
-        ]
-        assert len(dummyleaves) <= 1
-        for leaf in dummyleaves:
-            leaf.delete(prevent_nondicotomic=False)
-
         ctree = CollapsedTree(etetree)
 
+        # Fix internal node names to be unique, and verify
+        # The maps from nodes to names and nodes to sequences are bijections
+        n_nodes = 0
+        names = set()
+        seqs = set()
+        for node in ctree.tree.traverse():
+            n_nodes += 1
+            names.add(node.name)
+            seqs.add(node.sequence)
+        n_names, n_seqs = len(names), len(seqs)
+        if not (n_nodes == n_names and n_names == n_seqs):
+            raise RuntimeError(
+                "Multiple sequences with the same name, or multiple"
+                "names for the same sequence, observed in collapsed tree."
+            )
+
         # Here can do some validation on the tree:
         if self._validation_stats is not None:
             # root name:
-            if self._validation_stats["root"] not in ctree.tree.name:
+            if self._validation_stats["root"] != ctree.tree.name:
                 raise RuntimeError(
                     f"collapsed tree should have root name '{self._validation_stats['root']}' but has instead {ctree.tree.name}"
                 )
             # counts:
             counts = self._validation_stats["counts"]
             for node in etetree.iter_leaves():
-                if node.name:
-                    assert counts[node.name] == node.abundance
+                assert (
+                    sum(counts[og_id] for og_id in node.original_ids) == node.abundance
+                )
+                assert node.name in node.original_ids
+            tree_abundance = 0
             for node in ctree.tree.traverse():
-                if isinstance(node.name, tuple):
-                    for name in node.name:
-                        if name in counts:
-                            assert node.abundance == counts[name]
+                tree_abundance += node.abundance
+                if node.name in counts:
+                    assert (
+                        sum(counts[og_id] for og_id in node.original_ids)
+                        == node.abundance
+                    )
+                    assert node.name in node.original_ids
                 else:
-                    if node.name in counts:
-                        assert counts[node.name] == node.abundance
-                    else:
-                        assert node.abundance == 0
+                    assert node.abundance == 0
+            assert tree_abundance == sum(counts.values())
 
             # unnamed_seq issue:
             for node in ctree.tree.traverse():
@@ -1477,44 +1526,40 @@ class CollapsedForest:
                 raise RuntimeError(
                     "History DAG tree parsimony score does not match parsimony score provided"
                 )
-            # Root sequence:
-            if ctree.tree.sequence != self._validation_stats["root_seq"]:
+            # Root sequence is a possible disambiguation of root:
+            if any(
+                base not in gctree.utils.ambiguous_dna_values[ambig_base]
+                for base, ambig_base in zip(
+                    ctree.tree.sequence, self._validation_stats["root_seq"]
+                )
+            ):
                 raise RuntimeError(
-                    "History DAG root node sequence does not match root sequence provided"
+                    "History DAG root node sequence does not match root sequence provided\n"
+                    "found: " + ctree.tree.sequence + "\n"
+                    "expected: " + self._validation_stats["root_seq"]
                 )
             # Leaf names:
-            leaf_seqs = self._validation_stats["leaf_seqs"]
-            # A dictionary of leaf sequences to leaf names
-            observed_set = {
-                node for node in ctree.tree.iter_descendants() if node.abundance > 0
-            }
-            for node in observed_set:
-                if not node.is_root() and leaf_seqs[node.sequence] != node.name:
+            if "leaf_seqs" in self._validation_stats:
+                # Will be intentionally missing if observed sequences had
+                # ambiguities.
+                leaf_seqs = self._validation_stats["leaf_seqs"]
+                # A dictionary of leaf sequences to leaf names
+                observed_set = {
+                    node for node in ctree.tree.iter_descendants() if node.abundance > 0
+                }
+                for node in observed_set:
+                    if not node.is_root() and leaf_seqs[node.sequence] != node.name:
+                        raise RuntimeError(
+                            "History DAG tree leaf names don't match sequences"
+                        )
+                observed_seqs = {node.sequence for node in observed_set}
+                nonroot_observed_seqs = observed_seqs - {ctree.tree.sequence}
+                nonroot_leaf_seqs = set(leaf_seqs.keys()) - {ctree.tree.sequence}
+                if nonroot_leaf_seqs != nonroot_observed_seqs:
                     raise RuntimeError(
-                        "History DAG tree leaf names don't match sequences"
+                        "Observed nonroot sequences in history DAG tree don't match "
+                        "observed nonroot sequences passed in leaf_seqs."
                     )
-            observed_seqs = {node.sequence for node in observed_set}
-            nonroot_observed_seqs = observed_seqs - {ctree.tree.sequence}
-            nonroot_leaf_seqs = set(leaf_seqs.keys()) - {ctree.tree.sequence}
-            if nonroot_leaf_seqs != nonroot_observed_seqs:
-                raise RuntimeError(
-                    "Observed nonroot sequences in history DAG tree don't match "
-                    "observed nonroot sequences passed in leaf_seqs."
-                )
-            # The maps from nodes to names and nodes to sequences are bijections
-            n_nodes = 0
-            names = set()
-            seqs = set()
-            for node in ctree.tree.traverse():
-                n_nodes += 1
-                names.add(node.name)
-                seqs.add(node.sequence)
-            n_names, n_seqs = len(names), len(seqs)
-            if not (n_nodes == n_names and n_names == n_seqs):
-                raise RuntimeError(
-                    "Multiple sequences with the same name, or multiple"
-                    "names for the same sequence, observed in collapsed tree."
-                )
         else:
             warnings.warn("No validation was performed on tree")
 
@@ -1596,102 +1641,159 @@ def _lltree(cm_counts, p: np.float64, q: np.float64) -> Tuple[np.float64, np.nda
     return logf, grad_ll_genotype
 
 
-def _make_dag(trees, sequence_counts={}, from_copy=True):
+def _is_ambiguous(sequence):
+    return any(base not in gctree.utils.bases for base in sequence)
+
+
+def _make_dag(trees, from_copy=True):
     """Build a history DAG from ambiguous or disambiguated trees, whose nodes
     have abundance, name, and sequence attributes."""
     # preprocess trees so they're acceptable inputs
     # Assume all trees have fixed root sequence and fixed leaf sequences
-    def get_sequence(node):
-        if any(base not in gctree.utils.bases for base in node.sequence):
-            raise ValueError(
-                f"Unrecognized base found in node '{node.name}'. "
-                "Ambiguous bases are not permitted in observed sequences."
-            )
-        else:
-            return node.sequence
 
-    leaf_seqs = {get_sequence(node) for node in trees[0].get_leaves()}
-
-    sequence_counts = sequence_counts.copy()
     if from_copy:
         trees = [tree.copy() for tree in trees]
-    if all(len(tree.children) > 1 for tree in trees):
-        pass  # We're all good!
-    elif trees[0].sequence not in leaf_seqs:
-        if trees[0].sequence not in sequence_counts:
-            sequence_counts[trees[0].sequence] = 0
-        for tree in trees:
-            newleaf = tree.add_child(name="", dist=0)
-            newleaf.add_feature("sequence", trees[0].sequence)
-            if tree.sequence != newleaf.sequence:
-                raise ValueError(
-                    "At least some trees unifurcate at root, but root sequence is not fixed."
-                )
-    else:
-        # This should never happen in parsimony setting, when internal edges
-        # are collapsed by sequence
-        raise RuntimeError(
-            "Root sequence observed, but the corresponding leaf is not a child of the root node. "
-            "Gctree inference may give nonsensical results. Are you sure these are parsimony trees?"
-        )
 
-    dag = hdag.history_dag_from_etes(
-        trees,
-        ["sequence"],
-        attr_func=lambda n: {
-            "name": n.name,
-            "isotype": frozendict(),
-        },
-    )
-    # If there are too many ambiguities at too many nodes, disambiguation will
-    # hang. Need to have an alternative (disambiguate each tree before putting in dag):
-    if (
-        dag.count_trees(expand_count_func=hdag.utils.sequence_resolutions_count)
-        / dag.count_trees()
-        > 500000000
-    ):
+    # disambiguate leaves: disambiguate each tree and transplant disambiguated
+    # leaf sequences to tree with ambiguous internal sequences
+
+    # add pseudo-leaf below root in all trees:
+    # This is done first in case root is ambiguous, and gets merged with
+    # another ambiguous leaf node after disambiguation.
+    rootname = trees[0].name  # all root nodes must have the same name
+    for tree in trees:
+        newleaf = tree.add_child(name=tree.name, dist=0)
+        newleaf.add_feature("sequence", tree.sequence)
+        newleaf.add_feature("abundance", tree.abundance)
+
+    if any(_is_ambiguous(leaf.sequence) for leaf in trees[0].iter_leaves()):
         warnings.warn(
-            "Parsimony trees have too many ambiguities for disambiguation in history DAG. "
-            "Disambiguating trees individually. History DAG may find fewer new parsimony trees."
+            "Some observed sequences are ambiguous. A disambiguation consistent"
+            " with each dnapars tree will be chosen arbitrarily. Many alternative"
+            " disambiguated leaf sequences may be possible."
         )
-        distrees = [disambiguate(tree) for tree in trees]
-        dag = hdag.history_dag_from_etes(
-            distrees,
+        for tree in trees:
+            for node in tree.iter_leaves():
+                node.add_feature("original_ids", {node.name})
+            disambig_tree = tree.copy()
+            node_map = {
+                d_node: o_node
+                for d_node, o_node in zip(disambig_tree.traverse(), tree.traverse())
+            }
+            disambiguate(disambig_tree)
+
+            # remove duplicate leaves, and adjust abundances
+            leaf_seqs = {}
+            to_delete = []
+            for leaf in disambig_tree.iter_leaves():
+                if leaf.sequence in leaf_seqs:
+                    leaf_seqs[leaf.sequence].append(leaf)
+                else:
+                    leaf_seqs[leaf.sequence] = [leaf]
+            for sequence, leaf_list in leaf_seqs.items():
+                if len(leaf_list) > 1:
+                    # Always choose root pseudo-leaf to represent nodes, if
+                    # possible
+                    ancestor = disambig_tree.get_common_ancestor(*leaf_list)
+                    _leaf_list_names = {node.name: node for node in leaf_list}
+                    if rootname in _leaf_list_names:
+                        rep_node = _leaf_list_names.pop(rootname)
+                    else:
+                        rep_node = _leaf_list_names.pop(leaf_list[0].name)
+                    to_delete = list(_leaf_list_names.values())
+                    rep_node.abundance = sum(leaf.abundance for leaf in leaf_list)
+                    rep_node.original_ids = {
+                        seq_id for node in leaf_list for seq_id in node.original_ids
+                    }
+                    while to_delete:
+                        for node in to_delete:
+                            node_map[node].delete(prevent_nondicotomic=False)
+                            node.delete(prevent_nondicotomic=False)
+                        to_delete = [
+                            leaf
+                            for leaf in ancestor.iter_leaves()
+                            if (leaf.sequence == sequence and leaf != rep_node)
+                        ]
+            # transplant leaf sequences and abundances:
+            for node in disambig_tree.iter_leaves():
+                node_map[node].abundance = node.abundance
+                node_map[node].sequence = node.sequence
+                node_map[node].original_ids = node.original_ids
+            # remove some unifurcations
+            to_delete = []
+            for node in tree.iter_descendants():
+                if len(node.children) == 1:
+                    # this excludes leaves
+                    to_delete.append(node)
+            for node in to_delete:
+                node.delete(prevent_nondicotomic=False)
+
+    def trees_to_dag(trees):
+        return hdag.history_dag_from_etes(
+            trees,
             ["sequence"],
+            label_functions={"abundance": lambda n: n.abundance if n.is_leaf() else 0},
             attr_func=lambda n: {
                 "name": n.name,
+                "original_ids": (
+                    n.original_ids
+                    if "original_ids" in n.features
+                    else {n.name}
+                    if n.is_leaf()
+                    else set()
+                ),
                 "isotype": frozendict(),
             },
         )
+
+    dag = trees_to_dag(trees)
+    # If there are too many ambiguities at too many nodes, disambiguation will
+    # hang. Need to have an alternative (disambiguate each tree before putting in dag):
+
+    def test_explode_individually():
+        try:
+            if (
+                dag.count_trees(expand_count_func=hdag.utils.sequence_resolutions_count)
+                / dag.count_trees()
+                > 5000000
+            ):
+                return True
+            else:
+                return False
+        except OverflowError:
+            return True
+
+    if test_explode_individually():
+        warnings.warn(
+            "Parsimony trees have too many ambiguities for disambiguation in all possible ways. "
+            "Disambiguating trees individually. Gctree may find fewer parsimony trees."
+        )
+        distrees = [disambiguate(tree) for tree in trees]
+        dag = trees_to_dag(distrees)
     dag.explode_nodes(expand_func=hdag.utils.sequence_resolutions)
     # Look for (even) more trees:
     dag.add_all_allowed_edges(adjacent_labels=True)
     dag.trim_optimal_weight()
-    dag.convert_to_collapsed()
-    # Add abundances to attrs:
     for node in dag.preorder(skip_root=True):
-        if node.label.sequence in sequence_counts:
-            node.attr["abundance"] = sequence_counts[node.label.sequence]
-        else:
-            if node.is_leaf():
-                raise ValueError(
-                    "sequence_counts dictionary should contain all leaf sequences"
-                )
-            node.attr["abundance"] = 0
+        if node.label.abundance != 0 and not node.is_leaf():
+            raise RuntimeError(
+                f"An internal node {node.attr['name']} was found with nonzero abundance {node.label.abundance}"
+            )
+    dag.convert_to_collapsed()
+    dag.recompute_parents()
+    # Only leaf nodes have nonzero abundance now, so all internal edges are
+    # collapsed by sequence. Now labels with correct abundances are placed on
+    # leaf-adjacent nodes that will be collapsed by sequence.
+    for node in dag.preorder(skip_root=True):
+        if node.is_leaf():
+            for parent in node.parents:
+                if parent.label.sequence == node.label.sequence:
+                    parent.label = node.label
 
     if len(dag.hamming_parsimony_count()) > 1:
         raise RuntimeError(
             f"History DAG parsimony search resulted in parsimony trees of unexpected weights:\n {dag.hamming_parsimony_count()}"
         )
-    for node in dag.preorder(skip_root=True):
-        if (
-            node.attr["abundance"] != 0
-            and not node.is_leaf()
-            and frozenset({node.label}) not in node.clades
-        ):
-            raise RuntimeError(
-                "An internal node not adjacent to a leaf with the same label was found with nonzero abundance."
-            )
 
     # names on internal nodes are all messed up from disambiguation step, we'll
     # fix them in CollapsedTree.__init__.
@@ -1703,14 +1805,14 @@ def _cmcounter_dagfuncs():
     the DAG."""
 
     def edge_weight_func(n1, n2):
-        if n1.label == n2.label and n2.is_leaf():
+        if n2.is_leaf() and n1.label.sequence == n2.label.sequence:
             # Then this is a leaf-adjacent node with nonzero abundance
             return multiset.FrozenMultiset()
         else:
             m = len(n2.clades)
             if frozenset({n2.label}) in n2.clades:
                 m -= 1
-            return multiset.FrozenMultiset([(n2.attr["abundance"], m)])
+            return multiset.FrozenMultiset([(n2.label.abundance, m)])
 
     def accum_func(cmsetlist: List[multiset.FrozenMultiset]):
         st = multiset.FrozenMultiset()
@@ -1756,16 +1858,16 @@ def _ll_genotype_dagfuncs(p: np.float64, q: np.float64) -> hdag.utils.AddFuncDic
         collapsed, then 0.
 
         Expects DAG to have abundances added so that each node has
-        "abundance" key in attr dict.
+        abundance feature on label.
         """
-        if n2.is_leaf() and n2.label == n1.label:
+        if n2.is_leaf() and n2.label.sequence == n1.label.sequence:
             return hdag.utils.FloatState(0.0, state=Decimal(0.0))
         else:
             m = len(n2.clades)
             # Check if this edge should be collapsed, and reduce mutant descendants
             if frozenset({n2.label}) in n2.clades:
                 m -= 1
-            c = n2.attr["abundance"]
+            c = n2.label.abundance
             if n1.is_root() and c == 0 and m == 1:
                 # Add pseudocount for unobserved root unifurcation
                 c = 1
